@@ -16,7 +16,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-#ident "$Id: table.c,v 1.14 2003/05/19 20:55:10 nalin Exp $"
+#ident "$Id: table.c,v 1.15 2003/05/21 21:42:16 nalin Exp $"
 #include "../config.h"
 #include <sys/types.h>
 #include <assert.h>
@@ -707,7 +707,7 @@ print_array(GValueArray *array)
 			}
 		}
 		printf(")");
-		g_value_array_free(array);
+		_vte_matcher_free_params_array(array);
 	}
 }
 
