@@ -16,7 +16,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-#ident "$Id: vteapp.c,v 1.56 2003/05/01 19:56:13 nalin Exp $"
+#ident "$Id: vteapp.c,v 1.62 2003/05/22 19:14:49 nalin Exp $"
 
 #include "../config.h"
 
@@ -86,6 +86,13 @@ char_size_changed(GtkWidget *widget, guint width, guint height, gpointer data)
 				      GDK_HINT_MIN_SIZE);
 }
 
+static void
+deleted_and_quit(GtkWidget *widget, GdkEvent *event, gpointer data)
+{
+	gtk_widget_destroy(GTK_WIDGET(data));
+	gtk_main_quit();
+}
+
 static void
 destroy_and_quit(GtkWidget *widget, gpointer data)
 {
@@ -96,7 +103,6 @@ static void
 destroy_and_quit_eof(GtkWidget *widget, gpointer data)
 {
 	g_print("Detected EOF.\n");
-	destroy_and_quit(widget, data);
 }
 static void
 destroy_and_quit_exited(GtkWidget *widget, gpointer data)
@@ -360,7 +366,7 @@ main(int argc, char **argv)
 	const char *background = NULL;
 	gboolean transparent = FALSE, audible = TRUE, blink = TRUE,
 		 debug = FALSE, dingus = FALSE, geometry = TRUE, dbuffer = TRUE,
-		 console = FALSE, scroll = FALSE;
+		 console = FALSE, scroll = FALSE, keep = FALSE;
 	long lines = 100;
 	const char *message = "Launching interactive shell...\r\n";
 	const char *font = NULL;
@@ -384,6 +390,7 @@ main(int argc, char **argv)
 			    "[-f font] "
 			    "[-g] "
 			    "[-h] "
+			    "[-k] "
 			    "[-n] "
 			    "[-t terminaltype]\n";
 	back.red = back.green = back.blue = 0xffff;
@@ -418,7 +425,7 @@ main(int argc, char **argv)
 	g_assert(i < (g_list_length(args) + 2));
 
 	/* Parse some command-line options. */
-	while ((opt = getopt(argc, argv, "B:CDT2abc:df:ghn:st:w:-")) != -1) {
+	while ((opt = getopt(argc, argv, "B:CDT2abc:df:ghkn:st:w:-")) != -1) {
 		gboolean bail = FALSE;
 		switch (opt) {
 			case 'B':
@@ -454,6 +461,9 @@ main(int argc, char **argv)
 			case 'g':
 				geometry = !geometry;
 				break;
+			case 'k':
+				keep = !keep;
+				break;
 			case 'n':
 				lines = atol(optarg);
 				if (lines == 0) {
@@ -490,7 +500,7 @@ main(int argc, char **argv)
 	 * delete event to the quit function.. */
 	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
 	g_signal_connect(G_OBJECT(window), "delete_event",
-			 GTK_SIGNAL_FUNC(gtk_main_quit), NULL);
+			 GTK_SIGNAL_FUNC(deleted_and_quit), window);
 
 	/* Create a box to hold everything. */
 	hbox = gtk_hbox_new(0, FALSE);
@@ -498,7 +508,9 @@ main(int argc, char **argv)
 
 	/* Create the terminal widget and add it to the scrolling shell. */
 	widget = vte_terminal_new();
-	gtk_widget_set_double_buffered(widget, dbuffer);
+	if (!dbuffer) {
+		gtk_widget_set_double_buffered(widget, dbuffer);
+	}
 	gtk_box_pack_start(GTK_BOX(hbox), widget, TRUE, TRUE, 0);
 
 	/* Connect to the "char_size_changed" signal to set geometry hints
@@ -560,6 +572,7 @@ main(int argc, char **argv)
 
 	/* Set some defaults. */
 	vte_terminal_set_audible_bell(VTE_TERMINAL(widget), audible);
+	vte_terminal_set_visible_bell(VTE_TERMINAL(widget), !audible);
 	vte_terminal_set_cursor_blinks(VTE_TERMINAL(widget), blink);
 	vte_terminal_set_scroll_background(VTE_TERMINAL(widget), scroll);
 	vte_terminal_set_scroll_on_output(VTE_TERMINAL(widget), FALSE);
@@ -661,9 +674,20 @@ main(int argc, char **argv)
 	}
 
 	/* Go for it! */
+	g_object_add_weak_pointer(G_OBJECT(widget), (gpointer*)&widget);
+	g_object_add_weak_pointer(G_OBJECT(window), (gpointer*)&window);
 	gtk_widget_show_all(window);
 
 	gtk_main();
 
+	g_assert(widget == NULL);
+	g_assert(window == NULL);
+
+	if (keep) {
+		while (TRUE) {
+			sleep(60);
+		}
+	}
+
 	return 0;
 }
